"""Module that contains environments representing control tasks."""
from plangym.control.balloon import BalloonEnv  # noqa: E402
from plangym.control.box_2d import Box2DEnv  # noqa: E402
from plangym.control.classic_control import ClassicControl  # noqa: E402
from plangym.control.dm_control import DMControlEnv  # noqa: E402
from plangym.control.lunar_lander import LunarLander  # noqa: E402
